from typing import Optional, Sequence, Tuple, Union
import jax
import jax.numpy as jnp
import jax.random as jr
import equinox as eqx
from jaxtyping import PRNGKeyArray, Array, Float, Scalar, jaxtyped
from beartype import beartype as typechecker

from ._sde import SDE, _get_log_prob_fn, Time, TimeFn


def get_diffusion_fn(sigma_fn: Union[TimeFn, eqx.Module]) -> TimeFn:
    """ Get diffusion coefficient function for VE SDE: dx = sqrt(d[sigma^2(t)]/dt)dw """
    def _diffusion_fn(t: Time) -> Scalar:
        _, dsigmadt = jax.jvp(
            lambda t: jnp.square(sigma_fn(t)), 
            primals=(t,), 
            tangents=(jnp.ones_like(t),),
            has_aux=False
        )
        return jnp.sqrt(dsigmadt) # = sqrt(d[sigma^2(t)]/dt) ?
    return _diffusion_fn


class VESDE(SDE):
    sigma_fn: TimeFn | eqx.Module
    weight_fn: TimeFn 

    @jaxtyped(typechecker=typechecker)
    def __init__(
        self, 
        sigma_fn: TimeFn | eqx.Module, 
        weight_fn: Optional[TimeFn] = None, 
        dt: float = 0.1, 
        t0: float = 0., 
        t1: float = 1.
    ):
        """
            Construct a Variance Exploding SDE.

            dx = sqrt(d[sigma_fn(t) ** 2]/dt)

            Args:
                sigma: default variance value
                dt: timestep width
        """
        super().__init__(dt=dt, t0=t0, t1=t1)
        self.sigma_fn = sigma_fn
        self.weight_fn = weight_fn

    @jaxtyped(typechecker=typechecker)
    def sde(self, x: Float[Array, "..."], t: Time) -> Tuple[Float[Array, "..."], Scalar]:
        drift = jnp.zeros_like(x)
        _, dsigma2dt = jax.jvp(
            lambda t: jnp.square(self.sigma_fn(t)), 
            primals=(t,), 
            tangents=(jnp.ones_like(t),),
            has_aux=False
        )
        diffusion = jnp.sqrt(dsigma2dt)
        return drift, diffusion

    @jaxtyped(typechecker=typechecker)
    def marginal_prob(self, x: Float[Array, "..."], t: Time) -> Tuple[Float[Array, "..."], Scalar]:
        """ 
            SDE:
                dx = sqrt(d[sigma^2(t)]/dt) * dw
            sigma(t) = exp(t) for example
                x(t) ~ G[x(t)|x(0), [sigma^2(t) - sigma^2(0)] * I 

            x(t) ~ G[x(t)|x(0), [sigma^2(t) - sigma^2(0)] * I
        """
        std = jnp.sqrt(jnp.square(self.sigma_fn(t)) - jnp.square(self.sigma_fn(0.))) 
        return x, std 

    @jaxtyped(typechecker=typechecker)
    def weight(self, t: Time, likelihood_weight: bool = False) -> Scalar:
        if self.weight_fn is not None and not likelihood_weight:
            weight = self.weight_fn(t)
        else:
            if likelihood_weight:
                weight = jnp.square(self.sigma_fn(t))
            else:
                weight = jnp.square(self.sigma_fn(t)) # Same for likelihood weighting
        return weight

    def prior_sample(self, key: PRNGKeyArray, shape: Sequence[int]) -> Float[Array, "..."]:
        return jr.normal(key, shape) * self.sigma_fn(self.t1) 

    def prior_log_prob(self, z: Float[Array, "..."]) -> Scalar:
        return _get_log_prob_fn(scale=self.sigma_fn(self.t1))(z)


from typing import Sequence, Optional, Callable
import jax
import jax.numpy as jnp
import jax.random as jr
import einops
import equinox as eqx
from jaxtyping import Key, Array, Float, jaxtyped
from beartype import beartype as typechecker


class AdaLayerNorm(eqx.Module):
    net: eqx.Module
    ln: eqx.nn.LayerNorm
    data_shape: Sequence[int]

    def __init__(
        self, 
        data_shape: Sequence[int], 
        condition_dim: int, 
        *, 
        key: Key
    ):
        """
            Adaptive layer norm; generate scale and shift parameters from conditioning context.
        """
        data_dim = jnp.prod(jnp.asarray(data_shape)).item() # Ensure this isn't a jax array
        self.net = eqx.nn.Linear(condition_dim, data_dim * 2, key=key)
        # Don't use bias or scale since these will be learnable through the conditioning context
        self.ln = eqx.nn.LayerNorm(data_shape, use_bias=False, use_weight=False)
        self.data_shape = data_shape

    def __call__(self, x: Array, conditioning: Array) -> Array:
        # Compute scale and shift parameters from conditioning context
        scale_and_shift = jax.nn.gelu(self.net(conditioning))
        scale, shift = jnp.split(scale_and_shift, 2)
        scale = scale.reshape(self.data_shape)
        shift = shift.reshape(self.data_shape)
        # Apply layer norm
        x = self.ln(x)
        # Apply scale and shift (same scale, shift to all elements)
        x = x * (1. + scale) + shift
        return x


class MixerBlock(eqx.Module):
    patch_mixer: eqx.nn.MLP
    hidden_mixer: eqx.nn.MLP
    num_patches: int
    hidden_size: int
    norm1: eqx.nn.LayerNorm
    norm2: eqx.nn.LayerNorm
    a_dim: int = None

    def __init__(
        self, 
        num_patches: int, 
        hidden_size: int, 
        mix_patch_size: int, 
        mix_hidden_size: int, 
        context_dim: Optional[int] = None,
        *, 
        key: Key
    ):
        tkey, ckey = jr.split(key)
        self.patch_mixer = eqx.nn.MLP(
            num_patches, 
            num_patches, 
            mix_patch_size, 
            depth=1, 
            key=tkey
        )
        self.hidden_mixer = eqx.nn.MLP(
            hidden_size, 
            hidden_size, 
            mix_hidden_size, 
            depth=1, 
            key=ckey
        )
        if context_dim is not None:
            key1, key2 = jr.split(key)
            self.norm1 = AdaLayerNorm(
                (hidden_size, num_patches), context_dim, key=key1
            )
            self.norm2 = AdaLayerNorm(
                (num_patches, hidden_size), context_dim, key=key2
            )
        else:
            self.norm1 = eqx.nn.LayerNorm((hidden_size, num_patches))
            self.norm2 = eqx.nn.LayerNorm((num_patches, hidden_size))
        self.hidden_size = hidden_size
        self.num_patches = num_patches 
        self.a_dim = context_dim

    def __call__(self, y: Array, a: Array = None) -> Array:
        if a is not None and self.a_dim is not None:
            y = y + jax.vmap(self.patch_mixer)(self.norm1(y, a))
        else:
            y = y + jax.vmap(self.norm1)(y)
        y = einops.rearrange(y, "c p -> p c")
        if a is not None and self.a_dim is not None:
            y = y + jax.vmap(self.hidden_mixer)(self.norm2(y, a))
        else:
            y = y + jax.vmap(self.norm2)(y)
        y = einops.rearrange(y, "p c -> c p")
        return y


def get_timestep_embedding(timesteps, embedding_dim):
    """Build sinusoidal embeddings (from Fairseq)."""
    # Convert scalar timesteps to an array
    assert embedding_dim % 2 == 0
    if jnp.isscalar(timesteps):
        timesteps = jnp.array(timesteps)
    timesteps *= 1000.
    half_dim = embedding_dim // 2
    emb = jnp.log(10_000.) / (half_dim - 1.)
    emb = jnp.exp(jnp.arange(half_dim) * -emb)
    emb = timesteps * emb
    emb = jnp.concatenate([jnp.sin(emb), jnp.cos(emb)], axis=0)
    return emb


def get_activation_fn(fn):
    if isinstance(fn, str):
        return getattr(jax.nn, fn)
    else:
        return fn


class Mixer2d(eqx.Module):
    conv_in: eqx.nn.Conv2d
    conv_out: eqx.nn.ConvTranspose2d
    blocks: list[MixerBlock]
    norm: eqx.nn.LayerNorm
    t1: float
    embedding_dim: int
    final_activation: callable
    img_size: Sequence[int]
    q_dim: int
    a_dim: int

    @jaxtyped(typechecker=typechecker)
    def __init__(
        self,
        img_size: Sequence[int],
        patch_size: int,
        hidden_size: int,
        mix_patch_size: int,
        mix_hidden_size: int,
        num_blocks: int,
        t1: float,
        embedding_dim: int = 8,
        final_activation: Optional[Callable | str] = None,
        q_dim: Optional[int] = None,
        a_dim: Optional[int] = None,
        *,
        key: Key[jnp.ndarray, "..."]
    ):
        """
            A 2D MLP Mixer model.
            This model processes 2D images using a patch-based approach, where 
            each patch is processed through a series of mixer blocks. It also 
            supports optional conditioning through `q_dim` and `a_dim`.

            Parameters:
            -----------
            `img_size` : `Sequence[int]`
                Shape of the input image as a sequence (typically height and width).
            
            `patch_size` : `int`
                Size of the patches into which the input image is divided.
            
            `hidden_size` : `int`
                Size of the hidden layers within each Mixer block.
            
            `mix_patch_size` : `int`
                Size of the patches used for the patch mixing within each Mixer block.
            
            `mix_hidden_size` : `int`
                Size of the hidden layers used for mixing channels across patches.
            
            `num_blocks` : `int`
                Number of Mixer blocks in the model.
            
            `t1` : `float`
                Maximum time of diffusion process. Scales input times.
            
            `embedding_dim` : `int`, default: `4`
                Dimensionality of the time embedding. Defaults to `8`.

            `final_activation` : `Optional[Union[callable, str]]`, default: `None`
                Final activation function on output of model. Supply as a string
                of one of the attributes of `jax.nn` or a custom callable.
            
            `q_dim` : `Optional[int]`, default: `None`
                The number of channels in the conditioning map. Can be `None`.
                Must be same dimension as input `x` in `__call__`
            
            `a_dim` : `Optional[int]`, default: `None`
                The number of parameters in the conditioning. Can be `None`.
            
            `key` : `Key`
                JAX random key used for initialization.
        """

        input_size, height, width = img_size
        assert (height % patch_size) == 0
        assert (width % patch_size) == 0
        num_patches = (height // patch_size) * (width // patch_size)
        inkey, outkey, *bkeys = jr.split(key, 2 + num_blocks)

        _input_size = input_size + q_dim if q_dim is not None else input_size
        _context_dim = embedding_dim + a_dim if a_dim is not None else embedding_dim

        self.img_size = img_size
        self.q_dim = q_dim
        self.a_dim = a_dim

        self.conv_in = eqx.nn.Conv2d(
            _input_size, 
            hidden_size, 
            patch_size, 
            stride=patch_size, 
            key=inkey
        )
        self.conv_out = eqx.nn.ConvTranspose2d(
            hidden_size, 
            input_size, 
            patch_size, 
            stride=patch_size, 
            key=outkey
        )
        self.blocks = [
            MixerBlock(
                num_patches, 
                hidden_size, 
                mix_patch_size, 
                mix_hidden_size, 
                context_dim=_context_dim,
                key=bkey
            ) 
            for bkey in bkeys
        ]
        self.norm = eqx.nn.LayerNorm((hidden_size, num_patches))
        self.t1 = t1
        self.embedding_dim = embedding_dim
        self.final_activation = get_activation_fn(final_activation)

    @jaxtyped(typechecker=typechecker)
    def __call__(
        self, 
        t: float | Float[Array, ""], 
        y: Float[Array, "..."], 
        q: Optional[Float[Array, "{self.q_dim} ..."]] = None, 
        a: Optional[Float[Array, "{self.a_dim}"]] = None, 
        *, 
        key: Optional[Key[jnp.ndarray, "..."]] = None
    ) -> Float[Array, "..."]:
        _, height, width = y.shape
        t = jnp.atleast_1d(t / self.t1)
        t = get_timestep_embedding(t, embedding_dim=self.embedding_dim)
        if q is not None:
            yq = jnp.concatenate([y, q])
            _input = yq
        else:
            _input = y
        y = self.conv_in(_input)
        _, patch_height, patch_width = y.shape
        y = einops.rearrange(y, "c h w -> c (h w)")
        if a is not None:
            a = jnp.atleast_1d(a)
            at = jnp.concatenate([a, t])
            _input = at
        else:
            _input = t
        for block in self.blocks:
            y = block(y, _input)
        y = self.norm(y)
        y = einops.rearrange(
            y, "c (h w) -> c h w", h=patch_height, w=patch_width
        )
        y = self.conv_out(y)
        if self.final_activation is not None:
            y = self.final_activation(y)
        return y


from typing import Sequence, Tuple, Self, Callable, Optional
import jax
import jax.numpy as jnp
import equinox as eqx
from jaxtyping import Key, Array, Float, Scalar

Time = Scalar | float
TimeFn = Callable[[Time], Scalar]


def default_weight_fn(t, *, beta_integral=None, sigma_fn=None): 
    assert not ((beta_integral is not None) and (sigma_fn is not None))
    return 1. - jnp.exp(-beta_integral(t))


class SDE(eqx.Module):
    """
        Abstract base class for Stochastic Differential Equations (SDEs) used in 
        score-based generative modeling and related diffusion models.

        This class defines the required interface and provides base functionality for 
        forward and reverse-time SDEs, prior sampling, and log-probability computation. 
        The user should subclass `SDE` and implement the following methods:
            - `sde()`: returns drift and diffusion coefficients
            - `marginal_prob()`: returns the parameters of the marginal distribution at time t
            - `prior_sample()`: returns samples from the terminal distribution p_T(x)
            - `prior_log_prob()`: returns log-probabilities under p_T(x)
            - `weight()`: returns weighting for loss functions

        Attributes:
            dt (float): Time discretization step size.
            t0 (float): Start time of the diffusion process.
            t1 (float): End time of the diffusion process.
    """
    dt: float
    t0: float
    t1: float

    def __init__(self, dt: float = 0.01, t0: float = 0., t1: float = 1.):
        """
            Initialize the base SDE with time parameters.

            Args:
                dt (float): Time step for the SDE solver.
                t0 (float): Initial time of the process.
                t1 (float): Terminal time of the process.
        """
        super().__init__()
        self.t0 = t0
        self.t1 = t1
        self.dt = dt

    def sde(self, x: Array, t: Time) -> Tuple[Array, Array]:
        """
            Return the drift and diffusion coefficients f(x, t), g(t) of the SDE.

            Must be implemented by subclass.
        """
        ...

    def marginal_prob(self, x: Array, t: Time) -> Tuple[Array, Array]:
        """ Parameters to determine the marginal distribution of the SDE, $p_t(x)$. """
        ...

    def prior_sample(self, key: Key, shape: Sequence[int]) -> Array:
        """ 
            Generate one sample from the prior distribution, $p_T(x)$. 
        """
        ...

    def weight(self, t: Time, likelihood_weight: bool = False) -> Array:
        """
            Return the training loss weight at time t.

            Args:
                t (float or Array): Time value(s).
                likelihood_weight (bool): Whether to use likelihood weighting (optional).

            Returns:
                Array: Scalar or array of weights.
        """
        ...

    def prior_log_prob(self, z: Array) -> Array:
        """
            Compute log-density of the prior distribution.

            Useful for computing the log-likelihood via probability flow ODE.

            Args:
                z: latent code

            Returns:
                log probability density
        """
        ...

    def reverse(self, score_fn: eqx.Module, probability_flow: bool = False) -> Self:
        """
            Create the reverse-time SDE/ODE.

            Args:
                score_fn: A time-dependent score-based model that takes x and t and returns the score.
                probability_flow: If `True`, create the reverse-time ODE used for probability flow sampling.

            Returns:
                SDE: A subclass implementing the reverse-time SDE.
        """

        sde_fn = self.sde

        if hasattr(self, "beta_integral_fn"):
            _sde_fn = self.beta_integral_fn 
        if hasattr(self, "sigma_fn"):
            _sde_fn = self.sigma_fn

        _dt = self.dt
        _t0 = self.t0
        _t1 = self.t1

        # Build the class for the reverse-time SDE.
        class RSDE(self.__class__, SDE):
            probability_flow: bool

            def __init__(self):
                self.probability_flow = probability_flow
                super().__init__(_sde_fn, dt=_dt, t0=_t0, t1=_t1)

            def sde(
                self, 
                x: Float[Array, "..."], 
                t: Time, 
                q: Optional[Float[Array, "..."]] = None,
                a: Optional[Float[Array, "..."]] = None
            ) -> Tuple[Float[Array, "..."], Scalar]:
                """ 
                    Create the drift and diffusion functions for the reverse SDE/ODE. 
                    - forward time SDE:
                        dx = f(x, t) * dt + g(t) * dw
                    - reverse time SDE:
                        dx = [f(x, t) - g^2(t) * score(x, t)] * dt + g(t) * dw
                    - ode of SDE:
                        dx = [f(x, t) - 0.5 * g^2(t) * score(x, t)] * dt (ODE => No dw)
                """
                t = jnp.asarray(t)
                c = 0.5 if self.probability_flow else 1.
                drift, diffusion = sde_fn(x, t)
                score = score_fn(t, x, q, a)
                # Drift coefficient of reverse SDE and probability flow only different by a factor
                drift = drift - jnp.square(diffusion) * score * c
                # Set the diffusion function to zero for ODEs (dw=0)
                diffusion = 0. if self.probability_flow else diffusion
                return drift, diffusion

        return RSDE()


def _get_log_prob_fn(scale: float = 1.) -> Callable:
    def _log_prob_fn(z: Array) -> Array:
        return jax.scipy.stats.norm.logpdf(z, loc=0., scale=scale).sum()
    return _log_prob_fn


from typing import Sequence, Optional
from jaxtyping import Key
import numpy as np
from ml_collections import ConfigDict

from ._mixer import Mixer2d
from ._mlp import ResidualNetwork 
from ._unet import UNet
from ._dit import DiT


def get_model(
    model_key: Key, 
    model_type: str, 
    config: ConfigDict, 
    data_shape: Sequence[int], 
    context_shape: Optional[Sequence[int]] = None, 
    parameter_dim: Optional[int] = None
) -> Mixer2d | ResidualNetwork | UNet | DiT:
    """
        Get the model based on the specified type and configuration.

        Args:
            model_key: JAX random key for model initialization.
            model_type: Type of the model to create (e.g., "Mixer", "UNet", "mlp", "DiT").
            config: Configuration dictionary containing model parameters.
            data_shape: Shape of the input data (e.g. image dimensions, channels first).
            context_shape: Shape of the context map, if applicable.
            parameter_dim: Dimension of the additional conditioning.
        Returns:
            An initialized instance of the specified model type.

        Raises:
            ValueError: If the model type is not recognized.
    """

    # Grab channel assuming 'q' is a map like x
    if context_shape is not None:
        context_channels, *_ = context_shape.shape 
    else:
        context_channels = None

    if model_type not in ["Mixer", "UNet", "mlp", "DiT"]:
        raise ValueError(
            f"Model type {model_type} is not recognized. "
            "Choose from 'Mixer', 'UNet', 'mlp', or 'DiT'."
        )

    if model_type == "Mixer":
        model = Mixer2d(
            data_shape,
            patch_size=config.model.patch_size,
            hidden_size=config.model.hidden_size,
            mix_patch_size=config.model.mix_patch_size,
            mix_hidden_size=config.model.mix_hidden_size,
            num_blocks=config.model.num_blocks,
            t1=config.sde.t1,
            final_activation=config.model.final_activation,
            q_dim=context_channels,
            a_dim=parameter_dim,
            key=model_key
        )
    if model_type == "UNet":
        n_channels, *_ = data_shape
        model = UNet(
            dim=config.model.hidden_size,
            channels=n_channels,
            dim_mults=config.model.dim_mults,
            attn_heads=config.model.heads,
            attn_dim_head=config.model.dim_head,
            dropout=config.model.dropout_rate,
            learned_sinusoidal_cond=True,
            random_fourier_features=True,
            q_channels=context_channels,
            a_dim=parameter_dim,
            key=model_key
        )
    if model_type == "mlp":
        model = ResidualNetwork(
            in_size=np.prod(data_shape),
            width_size=config.model.width_size,
            depth=config.model.depth,
            activation=config.model.activation,
            dropout_p=config.model.dropout_p,
            q_dim=context_channels,
            a_dim=parameter_dim,
            t1=config.sde.t1,
            key=model_key
        )
    if model_type == "DiT":
        model = DiT(
            img_size=config.img_size,
            channels=config.n_channels,
            embed_dim=config.embed_dim,
            patch_size=config.patch_size,
            depth=config.depth,
            n_heads=config.n_heads,
            q_dim=context_channels, # Number of channels in conditioning map
            a_dim=parameter_dim,    # Number of parameters in power spectrum model
            key=model_key
        )

    return model


import os 
from typing import Tuple 
import jax
import jax.numpy as jnp
import jax.random as jr 
from jaxtyping import Key, Array
import numpy as np
import torch
from torchvision import transforms

from .utils import Scaler, ScalerDataset, TorchDataLoader, InMemoryDataLoader

DATA_DIR = os.environ.get("DATA_DIR", "")


class MapDataset(torch.utils.data.Dataset):
    def __init__(self, tensors, transform=None):
        # Tuple of (images, contexts, targets), turn them into tensors
        self.tensors = tuple(
            torch.as_tensor(tensor) for tensor in tensors
        )
        self.transform = transform
        assert all(
            self.tensors[0].size(0) == tensor.size(0) 
            for tensor in self.tensors
        )

    def __getitem__(self, index):
        x = self.tensors[0][index] # Fields
        a = self.tensors[1][index] # Parameters

        if self.transform:
            x = self.transform(x)

        return x, a

    def __len__(self):
        return self.tensors[0].size(0)


def get_quijote_data(n_pix: int) -> Tuple[Array, Array]:

    assert (DATA_DIR is not None) and (DATA_DIR != ""), (
        "Note: you must supply a path with the Quijote hypercube parameters and simulations."
    )

    X = np.load(os.path.join(DATA_DIR, "quijote_fields.npy"))[:, np.newaxis, ...]
    A = np.load(os.path.join(DATA_DIR, "quijote_parameters.npy"))

    dx = int(256 / n_pix)
    X = X.reshape((-1, 1, n_pix, dx, n_pix, dx)).mean(axis=(3, 5))

    return X, A


def get_quijote_labels() -> Array:
    Q = np.load(os.path.join(DATA_DIR, "quijote_parameters.npy"))
    return Q


def quijote(key, n_pix, split=0.9):
    key_train, key_valid = jr.split(key)

    data_shape = (1, n_pix, n_pix)
    context_shape = None #(1, n_pix, n_pix)
    parameter_dim = 5

    X, A = get_quijote_data(n_pix) 

    print("Quijote data:", X.shape, A.shape)

    # min = X.min()
    # max = X.max()
    # X = (X - min) / (max - min) # ... -> [0, 1]
    # X = 2.0 * (X - min) / (max - min) - 1.0 # ... -> [-1, 1]
    X = (X - X.mean()) / X.std()

    A = (A - A.min()) / (A.max() - A.min()) # ... -> [0, 1]

    n_train = int(split * len(X))

    # scaler = Scaler() # [0,1] -> [-1,1]

    # Don't have many maps
    train_dataloader = InMemoryDataLoader(
        X=X[:n_train], A=A[:n_train], key=key_train
    )
    valid_dataloader = InMemoryDataLoader(
        X=X[n_train:], A=A[n_train:], key=key_valid
    )

    def label_fn(key, n):
        A = get_quijote_labels() 
        ix = jr.choice(key, jnp.arange(len(A)), (n,))
        Q = None
        A = A[ix]
        return Q, A

    return ScalerDataset(
        name="quijote",
        train_dataloader=train_dataloader,
        valid_dataloader=valid_dataloader,
        data_shape=data_shape,
        context_shape=context_shape,
        parameter_dim=parameter_dim,
        process_fn=None, #scaler,
        label_fn=label_fn
    )


from __future__ import annotations

from functools import partial
from typing import Tuple, Optional, Optional, List
from functools import partial

import jax
import jax.numpy as jnp
import jax.random as jr 
import equinox as eqx
import equinox as eqx
from jaxtyping import Key, Array, Float, jaxtyped
from beartype import beartype as typechecker

from einops import einsum, rearrange, repeat
import einx


def identity(x):
    return x


def exists(v):
    return v is not None


def default(v, d):
    return v if exists(v) else d


def stop_grad(a):
    return jax.lax.stop_gradient(a)


def cast_tuple(t, length=1):
    return t if isinstance(t, tuple) else ((t,) * length)


def divisible_by(num, den):
    return (num % den) == 0


def key_split_allowing_none(key, n=2, i=None):
    if key is not None:
        if i is not None:
            key = jr.fold_in(key, i)
        keys = jr.split(key, n)
    else: 
        keys = [None] * n
    return keys


class Upsample(eqx.Module):
    conv: eqx.nn.Conv2d

    def __init__(
        self, 
        dim: int, 
        dim_out: Optional[int] = None, 
        *, 
        key: Key
    ):
        self.conv = eqx.nn.Conv2d(
            dim, 
            default(dim_out, dim), 
            kernel_size=3, 
            padding=1, 
            key=key
        )

    def __call__(self, x: Array) -> Array:
        c, h, w = x.shape
        x = jax.image.resize(
            x, shape=(c, h * 2, w * 2), method='bilinear'
        )
        x = self.conv(x)
        return x


class Downsample(eqx.Module):
    conv: eqx.nn.Conv2d

    def __init__(
        self, 
        dim: int, 
        dim_out: Optional[int] = None, 
        *, 
        key: Key
    ):
        self.conv = eqx.nn.Conv2d(
            dim * 4, default(dim_out, dim), kernel_size=1, key=key
        )

    def __call__(self, x: Array) -> Array:
        x = rearrange(
            x, 'c (h p1) (w p2) -> (c p1 p2) h w', p1=2, p2=2
        )
        x = self.conv(x)
        return x


class RMSNorm(eqx.Module):
    scale: float # Array, learnable?
    gamma: Array

    def __init__(self, dim: int):
        self.scale = dim ** 0.5 # This is a default scaling
        self.gamma = jnp.zeros((dim, 1, 1))

    def __call__(self, x: Array) -> Array:
        # Normalise x, shift and scale 
        return (x / jnp.linalg.norm(x, ord=2, axis=0)) * (self.gamma + 1) * self.scale 


class SinusoidalPosEmb(eqx.Module):
    dim: int
    theta: int

    def __init__(self, dim: int, theta: int = 10_000):
        self.dim = dim
        self.theta = theta

    def __call__(self, x: Array) -> Array:
        x = jnp.atleast_1d(x)
        half_dim = self.dim // 2
        emb = jnp.log(self.theta) / (half_dim - 1)
        emb = jnp.exp(jnp.arange(half_dim) * -emb)
        emb = einx.multiply('i, j -> i j', x, emb) # emb = x * emb
        emb = jnp.concatenate([jnp.sin(emb), jnp.cos(emb)])
        return emb


class RandomOrLearnedSinusoidalPosEmb(eqx.Module):
    random: bool
    weights: Array

    def __init__(self, dim: int, is_random: bool = False, *, key: Key):
        assert divisible_by(dim, 2)
        half_dim = dim // 2
        self.weights = jr.normal(key, (half_dim,))
        self.random = is_random

    def __call__(self, x: Array) -> Array:
        x = jnp.atleast_1d(x) 
        weights = jax.lax.stop_gradient(self.weights) if self.random else self.weights
        freqs = x * weights * 2. * jnp.pi 
        fouriered = jnp.concatenate([jnp.sin(freqs), jnp.cos(freqs)])
        fouriered = jnp.concatenate([x, fouriered])
        return fouriered


class Block(eqx.Module):
    proj: eqx.nn.Conv2d
    norm: RMSNorm
    dropout: eqx.nn.Dropout

    def __init__(
        self, 
        dim: int, 
        dim_out: int, 
        dropout: float = 0., 
        *, 
        key: Key
    ):
        self.proj = eqx.nn.Conv2d(
            dim, dim_out, kernel_size=3, padding=1, key=key
        )
        self.norm = RMSNorm(dim_out)
        self.dropout = eqx.nn.Dropout(dropout)

    def __call__(
        self, 
        x: Array, 
        scale_shift: Tuple[Array, Array] = None, 
        key: Optional[Key] = None
    ) -> Array:
        x = self.proj(x)
        x = self.norm(x)

        if exists(scale_shift):
            scale, shift = scale_shift
            x = x * (scale + 1.) + shift

        return self.dropout(jax.nn.silu(x), key=key)


class ResnetBlock(eqx.Module):
    mlp: eqx.nn.Linear
    block1: Block
    block2: Block
    res_conv: eqx.nn.Conv2d

    def __init__(
        self, 
        dim: int, 
        dim_out: int, 
        *, 
        time_emb_dim: int = None, 
        dropout: float = 0., 
        key: Key
    ):
        keys = jr.split(key, 4)
        self.mlp = eqx.nn.Linear(
            time_emb_dim, dim_out * 2, key=keys[0]
        ) if exists(time_emb_dim) else None
        self.block1 = Block(dim, dim_out, dropout=dropout, key=keys[1])
        self.block2 = Block(dim_out, dim_out, key=keys[2])
        self.res_conv = eqx.nn.Conv2d(
            dim, dim_out, kernel_size=1, key=keys[3]
        ) if dim != dim_out else eqx.nn.Identity()

    def __call__(
        self, 
        x: Array, 
        time_emb: Optional[Array] = None, 
        *, 
        key: Key 
    ) -> Array:
        keys = key_split_allowing_none(key)

        scale_shift = None
        if exists(self.mlp) and exists(time_emb):
            time_emb = self.mlp(jax.nn.silu(time_emb))
            time_emb = rearrange(time_emb, 'c -> c 1 1')
            scale_shift = jnp.split(time_emb, 2)

        h = self.block1(x, scale_shift=scale_shift, key=keys[0])

        h = self.block2(h, key=keys[1])

        return h + self.res_conv(x)


class LinearAttention(eqx.Module):
    scale: float
    heads: int
    norm1: RMSNorm
    mem_kv: Array
    to_qkv: eqx.nn.Conv2d
    conv: eqx.nn.Conv2d
    norm2: RMSNorm

    def __init__(
        self,
        dim: int,
        heads: int = 4,
        dim_head: int = 32,
        num_mem_kv: int = 4,
        *,
        key: Key
    ):
        keys = jr.split(key, 3)

        self.scale = dim_head ** -0.5
        self.heads = heads
        hidden_dim = dim_head * heads

        self.norm1 = RMSNorm(dim)

        self.mem_kv = jr.normal(keys[0], (2, heads, dim_head, num_mem_kv))
        self.to_qkv = eqx.nn.Conv2d(
            dim, hidden_dim * 3, kernel_size=1, use_bias=False, key=keys[1]
        )

        self.conv = eqx.nn.Conv2d(hidden_dim, dim, kernel_size=1, key=keys[2])
        self.norm2 = RMSNorm(dim)

    def __call__(self, x: Array) -> Array:
        c, h, w = x.shape

        x = self.norm1(x)

        qkv = jnp.split(self.to_qkv(x), 3)
        q, k, v = tuple(
            rearrange(t, '(h c) x y -> h c (x y)', h=self.heads) for t in qkv
        )

        mk, mv = tuple(repeat(t, 'h c n -> h c n') for t in self.mem_kv)
        k, v = map(partial(jnp.concatenate, axis=-1), ((mk, k), (mv, v)))

        q = jax.nn.softmax(q, axis=-2)
        k = jax.nn.softmax(k, axis=-1)

        q = q * self.scale

        context = einsum(k, v, 'h d n, h e n -> h d e')

        out = einsum(context, q, 'h d e, h d n -> h e n')
        out = rearrange(
            out, 'h c (x y) -> (h c) x y', h=self.heads, x=h, y=w
        )
        return self.norm2(self.conv(out))


class Attention(eqx.Module):
    scale: float
    heads: int
    norm: RMSNorm
    mem_kv: Array
    to_qkv: eqx.nn.Conv2d
    to_out: eqx.nn.Conv2d

    def __init__(
        self,
        dim: int,
        heads: int = 4,
        dim_head: int = 32,
        num_mem_kv: int = 4,
        flash: bool = False,
        *,
        key: Key
    ):
        keys = jr.split(key, 3)
        self.scale = dim_head ** -0.5
        self.heads = heads
        hidden_dim = dim_head * heads

        self.norm = RMSNorm(dim)

        self.mem_kv = jr.normal(keys[0], (2, heads, num_mem_kv, dim_head))
        self.to_qkv = eqx.nn.Conv2d(
            dim, hidden_dim * 3, kernel_size=1, use_bias=False, key=keys[1]
        )
        self.to_out = eqx.nn.Conv2d(
            hidden_dim, dim, kernel_size=1, use_bias=False, key=keys[2]
        )

    def __call__(self, x: Array) -> Array:
        c, h, w = x.shape

        x = self.norm(x)

        qkv = jnp.split(self.to_qkv(x), 3)
        q, k, v = map(lambda t: rearrange(t, '(h c) x y -> h (x y) c', h=self.heads), qkv)

        mk, mv = map(lambda t: repeat(t, 'h n d -> h n d'), self.mem_kv)
        k, v = map(partial(jnp.concatenate, axis=-2), ((mk, k), (mv, v)))

        q = q * self.scale
        sim = einsum(q, k, 'h i d, h j d -> h i j')

        attn = jax.nn.softmax(sim, axis=-1)
        out = einsum(attn, v, 'h i j, h j d -> h i d')

        out = rearrange(out, 'h (x y) d -> (h d) x y', x=h, y=w)
        return self.to_out(out)
 

class TimeMLP(eqx.Module):
    embed: RandomOrLearnedSinusoidalPosEmb | SinusoidalPosEmb
    layers: List[eqx.Module]

    def __init__(self, embed, fourier_dim, time_dim, a_dim, *, key):
        keys = jr.split(key)
        self.embed = embed 
        self.layers = [
            eqx.nn.Linear(
                fourier_dim + a_dim if a_dim is not None else fourier_dim, 
                time_dim, 
                key=keys[0]
            ),
            jax.nn.gelu,
            eqx.nn.Linear(
                time_dim + a_dim if a_dim is not None else time_dim, 
                time_dim, 
                key=keys[1]
            )
        ]

    def __call__(self, t, a):
        t = self.embed(t)
        t = self.layers[0](jnp.concatenate([t, a]) if a is not None else t)
        t = self.layers[1](t)
        t = self.layers[2](jnp.concatenate([t, a]) if a is not None else t)
        return t


class UNet(eqx.Module):
    channels: int
    init_conv: eqx.nn.Conv2d
    random_or_learned_sinusoidal_cond: bool
    time_mlp: List[eqx.Module]

    downs: List[eqx.Module]
    ups: List[eqx.Module]

    mid_block1: ResnetBlock
    mid_attn: Attention | LinearAttention
    mid_block2: ResnetBlock

    out_dim: int
    final_res_block: ResnetBlock
    final_conv: eqx.nn.Conv2d

    q_channels: int
    a_dim: int

    @jaxtyped(typechecker=typechecker)
    def __init__(
        self,
        dim: int,
        init_dim: Optional[int] = None,
        out_dim: Optional[int] = None,
        dim_mults: Sequence[int, ...] = (1, 2, 4, 8),
        channels: int = 1,
        q_channels: Optional[int] = None,
        a_dim: Optional[int] = None,
        learned_variance: bool = False,
        learned_sinusoidal_cond: bool = False,
        random_fourier_features: bool = False,
        learned_sinusoidal_dim: int = 16,
        sinusoidal_pos_emb_theta: int = 10_000,
        dropout: float = 0.,
        attn_dim_head: int = 32,
        attn_heads: int = 4,
        full_attn: bool = False , # Defaults to full attention only for inner most layer
        flash_attn: bool = False,
        *,
        key: Key[jnp.ndarray, "..."]
    ):
        key_modules, key_down, key_mid, key_up, key_final = jr.split(key, 5)
        keys = jr.split(key_modules, 3)

        # Determine dimensions
        self.channels = channels
        self.q_channels = q_channels
        self.a_dim = a_dim

        init_dim = default(init_dim, dim)
        self.init_conv = eqx.nn.Conv2d(
            channels + q_channels if q_channels is not None else channels, 
            init_dim, 
            kernel_size=7, 
            padding=3, 
            key=keys[0]
        )

        dims = [init_dim, *map(lambda m: dim * m, dim_mults)]
        in_out = list(zip(dims[:-1], dims[1:]))

        # Time embeddings
        time_dim = dim * 4

        self.random_or_learned_sinusoidal_cond = learned_sinusoidal_cond or random_fourier_features

        if self.random_or_learned_sinusoidal_cond:
            sinu_pos_emb = RandomOrLearnedSinusoidalPosEmb(
                learned_sinusoidal_dim, random_fourier_features, key=keys[1]
            )
            fourier_dim = learned_sinusoidal_dim + 1
        else:
            sinu_pos_emb = SinusoidalPosEmb(
                dim, theta=sinusoidal_pos_emb_theta
            )
            fourier_dim = dim

        self.time_mlp = TimeMLP(
            sinu_pos_emb, fourier_dim, time_dim, a_dim, key=keys[3]
        )

        # Attention
        if not full_attn:
            full_attn = (*((False,) * (len(dim_mults) - 1)), True)

        num_stages = len(dim_mults)
        full_attn  = cast_tuple(full_attn, num_stages)
        attn_heads = cast_tuple(attn_heads, num_stages)
        attn_dim_head = cast_tuple(attn_dim_head, num_stages)

        assert len(full_attn) == len(dim_mults)

        # Prepare blocks
        FullAttention = partial(Attention, flash=flash_attn)
        resnet_block = partial(ResnetBlock, time_emb_dim=time_dim, dropout=dropout)

        # Layers
        self.downs = []
        self.ups = []
        num_resolutions = len(in_out)

        # Downsampling layers 
        for ind, (
            (dim_in, dim_out), layer_full_attn, layer_attn_heads, layer_attn_dim_head
        ) in enumerate(
            zip(in_out, full_attn, attn_heads, attn_dim_head)
        ):
            keys = jr.split(jr.fold_in(key_down, ind), 4)

            is_last = ind >= (num_resolutions - 1)

            attn_class = FullAttention if layer_full_attn else LinearAttention

            self.downs.append(
                [
                    resnet_block(dim_in, dim_in, key=keys[0]),
                    resnet_block(dim_in, dim_in, key=keys[1]),
                    attn_class(
                        dim_in, 
                        dim_head=layer_attn_dim_head, 
                        heads=layer_attn_heads, 
                        key=keys[2]
                    ),
                    Downsample(dim_in, dim_out, key=keys[3]) 
                    if not is_last else 
                    eqx.nn.Conv2d(
                        dim_in, dim_out, kernel_size=3, padding=1, key=keys[3]
                    )
                ]
            )

        # Middle layers + attention
        mid_dim = dims[-1]
        keys = jr.split(key_mid, 3)
        self.mid_block1 = resnet_block(mid_dim, mid_dim, key=keys[0])
        self.mid_attn = FullAttention(
            mid_dim, 
            heads=attn_heads[-1], 
            dim_head=attn_dim_head[-1],
            key=keys[1]
        )
        self.mid_block2 = resnet_block(mid_dim, mid_dim, key=keys[2])

        # Upsampling layers + skip connections
        for ind, (
            (dim_in, dim_out), layer_full_attn, layer_attn_heads, layer_attn_dim_head
        ) in enumerate(
            zip(*map(reversed, (in_out, full_attn, attn_heads, attn_dim_head)))
        ):
            keys = jr.split(jr.fold_in(key_up, ind), 4)

            is_last = ind == (len(in_out) - 1)

            attn_class = FullAttention if layer_full_attn else LinearAttention

            self.ups.append(
                [
                    resnet_block(dim_out + dim_in, dim_out, key=keys[0]),
                    resnet_block(dim_out + dim_in, dim_out, key=keys[1]),
                    attn_class(
                        dim_out, 
                        dim_head=layer_attn_dim_head, 
                        heads=layer_attn_heads, 
                        key=keys[2]
                    ),
                    Upsample(dim_out, dim_in, key=keys[3]) 
                    if not is_last else 
                    eqx.nn.Conv2d(
                        dim_out, dim_in, kernel_size=3, padding=1, key=keys[3]
                    )
                ]
            )

        default_out_dim = channels * (1 if not learned_variance else 2)
        self.out_dim = default(out_dim, default_out_dim)

        keys = jr.split(key_final)
        self.final_res_block = resnet_block(init_dim * 2, init_dim, key=keys[0])
        self.final_conv = eqx.nn.Conv2d(
            init_dim + q_channels if q_channels is not None else init_dim, 
            self.out_dim, 
            kernel_size=1, 
            key=keys[1]
        )

    @property
    def downsample_factor(self):
        return 2 ** (len(self.downs) - 1)

    @jaxtyped(typechecker=typechecker)
    def __call__(
        self, 
        t: Float[Array, ""], 
        x: Float[Array, "{self.channels} _ _"], 
        q: Optional[Float[Array, "{self.q_channels} _ _"]] = None, 
        a: Optional[Float[Array, "{self.a_dim}"]] = None, 
        key: Optional[Key[jnp.ndarray, "..."]] = None
    ) -> Float[Array, "{self.channels} _ _"]:
        assert all(
            [divisible_by(d, self.downsample_factor) for d in x.shape[-2:]]
        ), (
            f"Input dimensions {x.shape[-2:]} need to be divisible"
            f" by {self.downsample_factor}, given the UNet"
        )

        key_down, key_mid, key_up, key_final = key_split_allowing_none(key, n=4) 

        x = self.init_conv(jnp.concatenate([x, q]) if q is not None else x)
        r = x.copy()

        t = self.time_mlp(t, a)

        h = []
        for i, (block1, block2, attn, downsample) in enumerate(self.downs):
            keys = key_split_allowing_none(key_down, i=i)

            x = block1(x, t, key=keys[0])
            h.append(x)

            x = block2(x, t, key=keys[1])
            x = attn(x) + x
            h.append(x)

            x = downsample(x)

        keys = key_split_allowing_none(key_mid)
        x = self.mid_block1(x, t, key=keys[0])
        x = self.mid_attn(x) + x
        x = self.mid_block2(x, t, key=keys[1])

        for i, (block1, block2, attn, upsample) in enumerate(self.ups):
            keys = key_split_allowing_none(key_up, i=i)

            x = jnp.concatenate([x, h.pop()])
            x = block1(x, t, key=keys[0])

            x = jnp.concatenate([x, h.pop()])
            x = block2(x, t, key=keys[1])
            x = attn(x) + x

            x = upsample(x)

        x = jnp.concatenate([x, r])
        x = self.final_res_block(x, t, key=key_final)

        return self.final_conv(jnp.concatenate([x, q]) if q is not None else x)


import os
import jax.random as jr 
import jax.numpy as jnp
from jaxtyping import Key
from torchvision import transforms, datasets

from .utils import Scaler, ScalerDataset, TorchDataLoader


def flowers(path: str, key: Key, n_pix: int) -> ScalerDataset:

    key_train, key_valid = jr.split(key)

    data_shape = (3, n_pix, n_pix)
    parameter_dim = 1
    n_classes = 102

    scaler = Scaler()

    train_transform = transforms.Compose(
        [
            transforms.Resize((n_pix, n_pix)),
            transforms.RandomCrop(n_pix, padding=4, padding_mode='reflect'),
            transforms.RandomHorizontalFlip(),
            transforms.RandomVerticalFlip(),
            transforms.ToTensor(),
            transforms.Lambda(scaler.forward)
        ]
    )
    valid_transform = transforms.Compose(
        [
            transforms.Resize((n_pix, n_pix)),
            transforms.RandomCrop(n_pix, padding=4, padding_mode='reflect'),
            transforms.ToTensor(),
            transforms.Lambda(scaler.forward)
        ]
    )

    train_dataset = datasets.Flowers102(
        os.path.join(path, "datasets/flowers/"), 
        split="train", 
        download=True, 
        transform=train_transform,
        target_transform=transforms.Lambda(lambda x: x.float())
    )
    valid_dataset = datasets.Flowers102(
        os.path.join(path, "datasets/flowers/"), 
        split="val", 
        download=True, 
        transform=valid_transform,
        target_transform=transforms.Lambda(lambda x: x.float())
    )

    train_dataloader = TorchDataLoader(
        train_dataset, 
        data_shape=data_shape, 
        context_shape=None, 
        parameter_dim=parameter_dim,
        key=key_train
    )
    valid_dataloader = TorchDataLoader(
        valid_dataset, 
        data_shape=data_shape, 
        context_shape=None, 
        parameter_dim=parameter_dim,
        key=key_valid
    )

    def label_fn(key, n):
        Q = None
        A = jr.choice(key, jnp.arange(n_classes), (n,))
        A = A[:, jnp.newaxis].astype(jnp.float32)
        return Q, A

    return ScalerDataset(
        name="flowers",
        train_dataloader=train_dataloader,
        valid_dataloader=valid_dataloader,
        data_shape=data_shape,
        context_shape=None,
        parameter_dim=parameter_dim,
        process_fn=scaler,
        label_fn=label_fn
    )



import os
import jax.random as jr 
import jax.numpy as jnp
from jaxtyping import Key, Float, Array
from torch import Tensor
from torchvision import datasets, transforms

from .utils import Normer, Scaler, ScalerDataset, TorchDataLoader, InMemoryDataLoader


def tensor_to_array(tensor: Tensor) -> Array:
    return jnp.asarray(tensor.numpy())


def mnist(path:str, key: Key, *, in_memory: bool = True) -> ScalerDataset:

    key_train, key_valid = jr.split(key)

    n_pix = 28
    data_shape = (1, n_pix, n_pix)
    parameter_dim = 1 
    n_classes = 10

    train_transform = transforms.Compose(
        [
            transforms.Resize((n_pix, n_pix)),
            transforms.ToTensor() 
        ]
    )
    valid_transform = transforms.Compose(
        [
            transforms.Resize((n_pix, n_pix)),
            transforms.ToTensor()
        ]
    )

    # MNIST is small enough that the whole dataset can be placed in memory, so
    # we can actually use a faster method of data loading.
    train_dataset = datasets.MNIST(
        os.path.join(path, "datasets/mnist/"), 
        train=True, 
        download=True, 
        transform=train_transform,
        target_transform=transforms.Lambda(lambda x: x.float())
    )
    valid_dataset = datasets.MNIST(
        os.path.join(path, "datasets/mnist/"), 
        train=False, 
        download=True, 
        transform=valid_transform,
        target_transform=transforms.Lambda(lambda x: x.float())
    )

    if in_memory:
        train_data = tensor_to_array(train_dataset.data)[:, jnp.newaxis, ...] / 255.
        train_targets = tensor_to_array(train_dataset.targets)[:, jnp.newaxis]
        valid_data = tensor_to_array(valid_dataset.data)[:, jnp.newaxis, ...] / 255.
        valid_targets = tensor_to_array(valid_dataset.targets)[:, jnp.newaxis]

        train_targets = train_targets.astype(jnp.float32)
        valid_targets = valid_targets.astype(jnp.float32)

        mu, std = train_data.mean(), train_data.std()
        train_data = (train_data - mu) / std
        valid_data = (valid_data - mu) / std

        train_dataloader = InMemoryDataLoader(
            train_data, Q=None, A=train_targets, key=key_train
        )
        valid_dataloader = InMemoryDataLoader(
            valid_data, Q=None, A=valid_targets, key=key_valid
        )
    else:
        train_dataloader = TorchDataLoader(
            train_dataset, data_shape, context_shape=None, parameter_dim=parameter_dim, key=key_train
        )
        valid_dataloader = TorchDataLoader(
            valid_dataset, data_shape, context_shape=None, parameter_dim=parameter_dim, key=key_valid
        )


    def label_fn(key: Key, n: int) -> tuple[None, Float[Array, "n 1"]]:
        Q = None
        A = jr.choice(key, jnp.arange(n_classes), (n,))
        A = A[:, jnp.newaxis].astype(jnp.float32)
        return Q, A

    return ScalerDataset(
        name="mnist",
        train_dataloader=train_dataloader,
        valid_dataloader=valid_dataloader,
        data_shape=data_shape,
        context_shape=None,
        parameter_dim=parameter_dim,
        process_fn=None,
        label_fn=label_fn
    )


from typing import Tuple, Optional
import jax


def get_shardings() -> Tuple[
    Optional[jax.sharding.NamedSharding], Optional[jax.sharding.NamedSharding]
]:
    """
        Create sharding strategies based on the number of available JAX devices.

        If multiple devices are available, this function creates a 1D device mesh and returns:
        - a data-parallel sharding along the mesh dimension `'x'`
        - a fully replicated sharding across devices (for sharding a model across the devices)

        If only one device is available, both sharding strategies are returned as `None`.

        Returns
        -------
        sharding : Optional[jax.sharding.NamedSharding]
            Sharding strategy that partitions data across devices along axis `'x'`,
            or `None` if only one device is available.
        replicated : Optional[jax.sharding.NamedSharding]
            Fully replicated sharding strategy (no partitioning),
            or `None` if only one device is available.

        Notes
        -----
        - This function prints out the number and identities of local devices.
    """
    devices = jax.devices()
    n_devices = len(devices)

    print(f"Running on {n_devices} local devices: \n\t{devices}")

    if n_devices > 1:

        mesh = jax.sharding.Mesh(devices, "x")

        replicated = jax.sharding.NamedSharding(
            mesh, jax.sharding.PartitionSpec()
        )
        sharding = jax.sharding.NamedSharding(
            mesh, jax.sharding.PartitionSpec("x")
        )

    else:
        sharding = replicated = None

    return sharding, replicated